import Data.Map (Map)
import qualified Data.Map as Map

-- An evaluator that handles failure explicitly

-- pure lambda calculus
data Expr = 
    Var String       -- x
  | Lam String Expr  -- \ x. e
  | App Expr Expr    -- e1 e2
  | Number Integer   -- n
  | Add Expr Expr    -- e1 + e2
  deriving (Show)

data Val =
    Clo String Expr Env  -- \ x . e in env
  | NumberV Integer      -- n
  deriving (Show)

-- this is a type alias
type Env = Map String Val

-- Map.lookup :: (Eq k) => k -> Map k v -> Maybe v
-- Map.insert :: (Eq k) => k -> v -> Map k v -> Map k v
-- a big step interpreter
eval :: Expr -> Env -> Maybe Val
eval (Var x) env = Map.lookup x env
eval (Lam x body) env = Just (Clo x body env)
eval (App e1 e2) env =
  case (eval e1 env, eval e2 env) of
    (Just (Clo x body env'), Just v2) ->
      eval body (Map.insert x v2 env')
    _ -> Nothing
eval (Number n) env = Just (NumberV n)
eval (Add e1 e2) env = 
  case (eval e1 env, eval e2 env) of
    (Just (NumberV n1), Just (NumberV n2)) ->
      Just (NumberV (n1 + n2))
    _ -> Nothing

-- if you 'runhaskell' this file it will print 'oh hai' signaling that this
-- file type-checks
main :: IO ()
main = putStrLn "oh hai"
